unit sdvm86task;
interface

{$I sd_defines.inc}


uses gentypes,sdconsts, pmode;

const
  gVM86_Code:array[0..3] of Byte=($CD, $F5, $EB, $FC); // see vm86_a.asm
  // - this code invokes GPF to abandon VM86 task.


type
  PVM86Regs=^TVM86Regs;
  TVM86Regs=
  record
    eax,
    ebx,
    ecx,
    edx,
    esi,
    edi,
    esp,
    ebp,
    ds,
    es,
    fs,
    gs:DWord;
    CF, // carry flag
    ZF:Boolean; // zero flag
  end;


// --------------- VARIABLES ------------------------


var

  gVM_TSS:TTSS32; // VM86

  gVM_R0_Stack:array[0..R0_STACK_SZ-1] of Byte;
  // - stack for SS0 when handling IRQ in VM86


  gIsOnVM86Int:Boolean; // when the kernel is asked to handle a user Int
                        // in VM86, e.g. BIOS call etc.
                        // and it will be in VM86 task until INT VM86_INTARG
                        // will be encountered.



// --------------- PROCEDURES ------------------------

  procedure InitVM86;

  procedure LockVM86IRQ( irq_n:TUInt ); // lock it to forbid enabling/disabling
  procedure UnlockVM86IRQ( irq_n:TUInt );

  function EnableVM86IRQ( irq_n:TUInt; enable:Boolean ):Boolean;
  // - returns false if locked.


//  procedure CallInt86( int_n:TUInt; p_inout_regs:PVM86Regs );
  // - simulates real mode interrupt using registers specified.



implementation
uses genutils, kutils, sdcommon, sdexcep, sdintr, sdcontexts;
type
  TVIRQList=set of 0..15;
var
  g_disabled_irqs:TVIRQList;
  g_locked_irqs:TVIRQList;
  
  g_exc_hnd_id, g_irq_hnd_id:TUInt;
  
  
  
  // -------------------
  // Emulates INT instruction, CS:IP must point beyond the INT instruction.
  // Must be used outside VM86 Context, modifies gVM_TSS !
  procedure emul_int86(intn:Byte);
  var
    p_stk:PWord; // - v86 stack pointer
  begin
    // get CS:IP pair, push FLAGS, CS, IP, clear IF, TF, AC
    p_stk:=PWord((gVM_TSS.ss shl 4)+gVM_TSS.esp);
    
    Dec(p_stk);
    p_stk^:=gVM_TSS.eflags and $FFFF;
    
    
    Dec(p_stk);
    p_stk^:=gVM_TSS.cs and $FFFF;
    
    
    Dec(p_stk);
    p_stk^:=gVM_TSS.eip and $FFFF;
    
    
    Dec(gVM_TSS.esp, 6);
    
    
    {$IFDEF DEBUG_VM86}
      
    KU_Print('int86 at CS:EIP (h)=');
    KU_PrintHex(gVM_TSS.cs);
    KU_Print(':');
    KU_PrintHex(gVM_TSS.eip);
    
    {$ENDIF}
      
        
    gVM_TSS.eflags:=gVM_TSS.eflags and 
      ( not (EFLAGS_IF or EFLAGS_TF or EFLAGS_AC));
      
    gVM_TSS.cs:=PWord(intn*4+2)^;
    gVM_TSS.eip:=PWord(intn*4)^;
    
    
    {$IFDEF DEBUG_VM86}
    
    KU_Print(' int86 to CS:EIP (h)=');
    KU_PrintHex(gVM_TSS.cs);
    KU_Print(':');
    KU_PrintHex(gVM_TSS.eip);
    KU_PrintLn('');
    
    {$ENDIF}
        
  end; // emul_int86

  
  // --------------------------------------------------
  // GENERAL PROTECTION Fault Handler for VM86
  // --------------------------------------------------
  
  
  procedure handle_vm86_GP;
  var
      p_op:PByte; // instruction caused exception
      i,rem:TUInt;
      
  label skp_transfer;    
  begin
    p_op:=Pointer(gVM_TSS.eip+(gVM_TSS.cs shl 4));
    //  GPF handling for VM86
    
    // klogmsg('switch');  
    
    // if it's an INT instruction then check its argument
    // VM86_INTARG.
    
    // 1) Check whether it is an INT xx instruction
    // 2) Check for INT*
    // 3) Check for CLI, STI, POPF, PUSHF, IRET.
    
    // see Intel Manual Vol. 2 for details - several checks must be implemented!!!
    //        for full compatibility!!!
    
    
    (*
      Opcode
      
      CC      INT 3       Interrupt 3 - trap to debugger
      CD ib   INT imm8    
      CE      INTO        Interrupt 4 - if overflow flag is 1 (EFLAGS_OF)
      
      ------    
      
      FA      CLI         Clear IF
      FB      STI         Set IF
      
      9D      POPF        POP from stack 16-bits of EFLAGS
      9C      PUSHF       PUSH 16-bits of EFLAGS into stack
      
      CF      IRET        pop IP, pop CS, pop FLAGS
    
    *)
    
    {$IFDEF DEBUG_VM86}
    
    KU_Print('VM86: Emulating instruction...');
    
    {$ENDIF}
    
    case p_op^ of
      $FA: begin  // CLI
             gVM_TSS.eflags:=gVM_TSS.eflags and (not EFLAGS_IF);
             Inc(gVM_TSS.eip);
             
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('CLI');
             {$ENDIF}
           end;
           
      $FB: begin // STI
             gVM_TSS.eflags:=gVM_TSS.eflags or EFLAGS_IF;
             Inc(gVM_TSS.eip);
             
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('STI');
             {$ENDIF}
           end;
  
      $9C: begin // PUSHF
             Dec(gVM_TSS.esp, 2);
             PWord(gVM_TSS.esp+(gVM_TSS.ss shl 4))^:=DWord(gVM_TSS.eflags 
               and $FFFF);
               
             Inc(gVM_TSS.eip);
             
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('PUSHF');
             {$ENDIF}
           end;
           
      $9D: begin // POPF           
             gVM_TSS.eflags:=(gVM_TSS.eflags and $FFFF0000) 
               or PWord(gVM_TSS.esp+(gVM_TSS.ss shl 4))^;
               
             Inc(gVM_TSS.esp, 2);
             Inc(gVM_TSS.eip);
             
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('POPF');
             {$ENDIF}
           end;
           
      $CF: begin // IRET
      
             gVM_TSS.eflags:=(gVM_TSS.eflags and $FFFF0000) 
               or PWord(gVM_TSS.esp+4+(gVM_TSS.ss shl 4))^;
               
             gVM_TSS.cs:=PWord(gVM_TSS.esp+2+(gVM_TSS.ss shl 4))^;
             
             gVM_TSS.eip:=PWord(gVM_TSS.esp+0+(gVM_TSS.ss shl 4))^;
               
             Inc(gVM_TSS.esp, 6);
             
             {$IFDEF DEBUG_VM86}
             
             KU_Print('IRET to CS:EIP (h)=');
             KU_PrintHex(gVM_TSS.cs);
             KU_Print(':');
             KU_PrintHex(gVM_TSS.eip);
             KU_PrintLn('');
             
             {$ENDIF}
           end;
           
      $CC: begin // INT 3
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('INT3');
             {$ENDIF}
             
             Inc(gVM_TSS.eip);
             emul_int86(3);
           end;
  
      $CE: begin // INTO
             {$IFDEF DEBUG_VM86}
             KU_PrintLn('INTO');
             {$ENDIF}
             
             Inc(gVM_TSS.eip);
             
             if (gVM_TSS.eflags and EFLAGS_OF) <>0 then
             begin
               emul_int86(4);
             end;
           end;
           
      $CD: begin // INT imm8
             {$IFDEF DEBUG_VM86}
             KU_Print('INT $');
             KU_PrintHex(p_op[1]);
             KU_PrintLn('');
             {$ENDIF}
      
             Inc(gVM_TSS.eip, 2);
             
             if p_op[1]=VM86_INTARG then // check for VM86 Exit
             begin
               PM_ChangeIRET_Task( @gGDT[VM_TSS_DESC_N], 
                                   @gExcTSS[13],
                                   SYS_TSS_DESC_N*8,
                                   @gGDT[SYS_TSS_DESC_N] );
                                   
               gIsOnVM86Int:=false; // signal VM86 BIOS call completion.
               goto skp_transfer;
             end;
             
             emul_int86( p_op[1] );
             
       skp_transfer:; // no transfer for virtual vm86->pm gate interrupt
             
             
           end;
             
    else // handle unknown opcode sequence
      begin
             
        KU_PrintLn('Emulation FAILED!');
        KU_Print('VM86_FATAL: Unk. opcode seq.= ');
        
        for i:=1 to 8 do
        begin
          KU_PrintHex(p_op^);
          Inc(p_op);
        end;
        
        DumpTSS( VM_TSS_DESC_N, @gVM_TSS);
        
                
        //i:=GU_Div( gVM_TSS.edx, 8, @rem);
        //i:=gVM_TSS.iomap[i];
        
        //KU_Print('IOMAP[]=');
        //KU_PrintHex(i);
        //KU_PrintLn('');
        
        StopKernel;
      end;
    end; // case
      
  end; // handle_vm86_GP


  //////// MATH
  
  procedure handle_vm86_NM;
  begin
    KU_PrintLn('Unimplemented math exception in VM86...');
    
    StopKernel;
  end;  
    

  function VM86ExcHandler( exc_num:TUint; ctx_id:TUInt ):Boolean;
  var
    p_ctx:PSDContext;
  begin
  
    p_ctx:=GetContextFromID( ctx_id );
  
    if p_ctx^.tss_tr.trval<>VM_TSS_DESC_N*8 then exit(false);
    
    case exc_num of  
      EXCEPTION_GP: handle_vm86_GP;
      EXCEPTION_NM: handle_vm86_NM; // math
        else
          begin
            KU_Print('VM86 FATAL: Unhandled Exception $');
            KU_PrintHex(exc_num);
            KU_PrintLn('!');
            StopKernel;
          end; // else
    end; // case excnum


    exit(true);
  end; // VM86ExcHandler



  procedure VM86IRQHandler( irq_n:TUInt; vm86:Boolean; p_frame:PZeroRingStackFrame );
  var
    int_num:Byte;
    p_stk16:PWord;
  begin
    if irq_n in g_disabled_irqs then exit;
    
    sdintr.gShouldSendEOI:=false;
    
    if irq_n<8 then int_num:=8+irq_n
      else int_num:=$70+irq_n;

  
    if not vm86 then
    begin
      emul_int86( int_num );
      
      {$IFDEF DEBUG_VM86}
          
      KU_PrintLn('Jumping to VM86 to emulate IRQ...');
      
      {$ENDIF}
      
      asm
        // FAR JUMP PTR16:32
        db $EA
        dd 0
        dw (VM_TSS_DESC_N*8)
    end; 
    end // if not vm86
      else // already in vm86
      begin
        Dec(p_frame^.esp, 6);
        // push flags, cs, ip
        
        p_stk16:=PWord(p_frame^.esp+(p_frame^.ss shl 4));
                
        p_stk16^:=Word(p_frame^.eip and $FFFF);                
        Inc(p_stk16);
        p_stk16^:=p_frame^.cs;        
        Inc(p_stk16);        
        p_stk16^:=Word(p_frame^.eflags and $FFFF);
        
        // set new flags, cs, ip
        p_frame^.eflags:=p_frame^.eflags and 
           ( not (EFLAGS_IF or EFLAGS_TF or EFLAGS_AC));
           
        p_frame^.eip:=PWord(int_num*4)^ and $FFFF;
        p_frame^.cs:=PWord(int_num*4+2)^ and $FFFF;
      end; // else
      
  end; // VM86IRQHandler


  procedure InitVM86;
  var p:Pointer;
  begin
    FillByte( gVM_TSS, sizeof(gVM_TSS), 0 );

    gVM_TSS.eflags:=CLEAR_EFLAGS or EFLAGS_VM; //VM, IOPL=0
    gVM_TSS.esp0:=DWord(@gVM_R0_Stack[R0_STACK_SZ-5]);
    gVM_TSS.ss0:=16;

    gVM_TSS.cs:=$7C0;
    gVM_TSS.eip:=0;
    gVM_TSS.ss:=$7C0;
    gVM_TSS.esp:=$FFF0;
    gVM_TSS.ds:=$7C0;
    gVM_TSS.es:=$7C0;
    gVM_TSS.fs:=$7C0;
    gVM_TSS.gs:=$7C0;


    p:=Pointer($7C00);

    Move(gVM86_Code, p^, sizeof(gVM86_Code));


    PM_SetDescr(@gGDT[VM_TSS_DESC_N], DWord(@gVM_TSS), sizeof(TTSS32)-1,
      SEGTYPE_TSS32 or SEGTYPE_PRESENT,0);


    // IO bitmap:

    gVM_TSS.iomapoffs:=sizeof(TTSS32_0);//TUInt(@gVM_TSS.iomap)-TUInt(@gVM_TSS);
  //  FillByte(gVM_TSS.iomap, sizeof(gVM_TSS.iomap), $FF);

    gIsOnVM86Int:=false;
    
    g_disabled_irqs:=[];
    g_locked_irqs:=[];
    
    if not AddUserExcHandler( @VM86ExcHandler, g_exc_hnd_id ) then
    begin
      KU_PrintLn('FATAL: can''t add VM86ExcHandler!!!');
      
      StopKernel;
    end;

    
    if not AddUserIRQHandler( @VM86IRQHandler, g_irq_hnd_id ) then
    begin
      KU_PrintLn('FATAL: can''t add VM86IRQHandler!!!');
      
      StopKernel;
    end;
    
  end; // InitVM86
  
  
  
  procedure LockVM86IRQ( irq_n:TUInt );
  begin
    g_locked_irqs:=g_locked_irqs+[irq_n];
  end;
  
  
  procedure UnlockVM86IRQ( irq_n:TUInt );
  begin
    g_locked_irqs:=g_locked_irqs-[irq_n];
  end;

  
  function EnableVM86IRQ( irq_n:TUInt; enable:Boolean ):Boolean;
  begin
    if irq_n in g_locked_irqs then exit(false);
    
    if enable then 
      g_disabled_irqs:=g_disabled_irqs-[irq_n]
        else
          g_disabled_irqs:=g_disabled_irqs+[irq_n];
    
    exit(true);
  end;
  


end.